""" Module that scrapes a web page for hyperlinks """
import re
import six
import collections
import pprint

import requests

from bs4 import BeautifulSoup

content_allow = ['text/plain', 'text/html']


class Links(object):

    """Grabs links from a web page
    based upon a URL and filters"""

    def __init__(self, href=None, text=None, parser='lxml'):
        """ Create instance of Links class

        :param href: URL to download links from
        :param text: Search through text for links instead of URL
        :param parser: Parser for BeautifulSoup library. [default: lxml]
        """
        if href is not None and not href.startswith('http'):
            raise ValueError("URL must contain http:// or https://")
        elif href is not None:
            self._href = href
            self._response = requests.get(self._href)
            if self._response.status_code != 200:
                raise ValueError('Response FAIL: %s' %
                                 self._response.status_code)

            if not hasattr(self._response, 'headers'):
                raise ValueError('Response has not headers')

            if 'content-type' not in self._response.headers:
                raise ValueError('content-type information is not available')

            self._ctype = self._response.headers['content-type'].rsplit(';')[0]
            if self._ctype not in content_allow:
                raise ValueError(
                    'the content-type %s is not textable' % self._ctype)

            self._text = self._response.text

        elif href is None and text is not None:
            self._text = text
        else:
            raise ValueError("Either href or text must not be empty")

        self._soup = BeautifulSoup(self._text, parser)

    def __repr__(self):
        return "<Links {0}>".format(self._href or self._text[:15] + '...')

    @property
    def text(self):
        return self._text

    @property
    def response(self):
        return self._response

    def find(self, limit=None, reverse=False, sort=None,
             exclude=None, duplicates=True, pretty=False, **filters):
        """ Using filters and sorts, this finds all hyperlinks
        on a web page

        :param limit: Crop results down to limit specified
        :param reverse: Reverse the list of links, useful for before limiting
        :param exclude: Remove links from list
        :param duplicates: Determines if identical URLs should be displayed
        :param pretty: Quick and pretty formatting using pprint
        :param filters: All the links to search for """
        if exclude is None:
            exclude = []

        if filters is None:
            filters = {}
        search = self._soup.findAll('a', **filters)

        if reverse:
            search.reverse()

        links = []
        for anchor in search:
            build_link = anchor.attrs
            try:
                build_link[u'seo'] = seoify_hyperlink(anchor['href'])
            except KeyError:
                pass

            try:
                build_link[u'text'] = anchor.string or build_link['seo']
            except KeyError:
                pass

            ignore_link = False
            for nixd in exclude:
                for key, value in six.iteritems(nixd):
                    if key in build_link:
                        if (isinstance(build_link[key], collections.Iterable)
                                and not isinstance(build_link[key], six.string_types)):
                           for item in build_link[key]:
                                ignore_link = exclude_match(value, item)
                        else:
                            ignore_link = exclude_match(value, build_link[key])

            if not duplicates:
                for link in links:
                    if link['href'] == anchor['href']:
                        ignore_link = True

            if not ignore_link:
                links.append(build_link)

            if limit is not None and len(links) == limit:
                break

        if sort is not None:
            links = sorted(links, key=sort, reverse=reverse)

        if pretty:
            pp = pprint.PrettyPrinter(indent=4)
            return pp.pprint(links)
        else:
            return links


def exclude_match(exclude, link_value):
    """ Check excluded value against the link's current value """
    if hasattr(exclude, "search") and exclude.search(link_value):
        return True

    if exclude == link_value:
        return True

    return False


def seoify_hyperlink(hyperlink):
    """Modify a hyperlink to make it SEO-friendly by replacing
    hyphens with spaces and trimming multiple spaces.

    :param hyperlink: URL to attempt to grab SEO from """
    last_slash = hyperlink.rfind('/')
    return re.sub(r' +|-', ' ', hyperlink[last_slash + 1:])




""" Test data for test HTML pages """
pages = [{
    "url": "http://www.google.com",
    "num_links": 20,
    "limit_find": [{'seo': u'imghp?hl=en&tab=wi', 'text': u'Images', u'href': u'http://www.google.com/imghp?hl=en&tab=wi', u'class': [u'gb1']}, {'seo': u'maps?hl=en&tab=wl', 'text': u'Maps', u'href': u'http://maps.google.com/maps?hl=en&tab=wl', u'class': [u'gb1']}, {'seo': u'?hl=en&tab=w8', 'text': u'Play', u'href': u'https://play.google.com/?hl=en&tab=w8', u'class': [u'gb1']}, {'seo': u'?tab=w1', 'text': u'YouTube', u'href': u'http://www.youtube.com/?tab=w1', u'class': [u'gb1']}, {'seo': u'nwshp?hl=en&tab=wn', 'text': u'News', u'href': u'http://news.google.com/nwshp?hl=en&tab=wn', u'class': [u'gb1']}],
    "limit_reverse_find": [{'href': '/intl/en/policies/', 'seo': '', 'text': 'Privacy & Terms'}, {'href': '/intl/en/about.html', 'seo': 'about.html', 'text': 'About Google'}, {'href': 'https://plus.google.com/116899029375914044550', 'seo': '116899029375914044550', 'text': '+Google', 'rel': ['publisher']}, {'href': '/services/', 'seo': '', 'text': 'Business Solutions'}, {u'href': u'/intl/en/ads/', 'seo': '', 'text': u'Advertising\xa0Programs'}],
    "limit_sort_text": [{'seo': u'imghp?hl=en&tab=wi', 'text': u'Images', u'href': u'http://www.google.com/imghp?hl=en&tab=wi', u'class': [u'gb1']}, {'seo': u'maps?hl=en&tab=wl', 'text': u'Maps', u'href': u'http://maps.google.com/maps?hl=en&tab=wl', u'class': [u'gb1']}, {'seo': u'nwshp?hl=en&tab=wn', 'text': u'News', u'href': u'http://news.google.com/nwshp?hl=en&tab=wn', u'class': [u'gb1']}, {'seo': u'?hl=en&tab=w8', 'text': u'Play', u'href': u'https://play.google.com/?hl=en&tab=w8', u'class': [u'gb1']}, {'seo': u'?tab=w1', 'text': u'YouTube', u'href': u'http://www.youtube.com/?tab=w1', u'class': [u'gb1']}],
    "limit_sort_href": [{'seo': u'maps?hl=en&tab=wl', 'text': u'Maps', u'href': u'http://maps.google.com/maps?hl=en&tab=wl', u'class': [u'gb1']}, {'seo': u'nwshp?hl=en&tab=wn', 'text': u'News', u'href': u'http://news.google.com/nwshp?hl=en&tab=wn', u'class': [u'gb1']}, {'seo': u'imghp?hl=en&tab=wi', 'text': u'Images', u'href': u'http://www.google.com/imghp?hl=en&tab=wi', u'class': [u'gb1']}, {'seo': u'?tab=w1', 'text': u'YouTube', u'href': u'http://www.youtube.com/?tab=w1', u'class': [u'gb1']}, {'seo': u'?hl=en&tab=w8', 'text': u'Play', u'href': u'https://play.google.com/?hl=en&tab=w8', u'class': [u'gb1']}],
    "exclude_links": 12,
    "duplicate_links": 20
}, {
    "url": "http://www.freep.com/section/NLETTER10",
    "num_links": 144,
    "limit_find": [{'seo': u'm.freep.com', 'text': u'click here', u'style': u'color:#D19D4E;', u'href': u'http://m.freep.com'}, {'seo': u'%%=RedirectTo(@fullurl1)=%%', 'text': u'click here', u'style': u'color:#D19D4E;', u'href': u'%%=RedirectTo(@fullurl1)=%%', u'target': u'_blank'}, {'seo': u'www.freep.com', 'text': u'www.freep.com', u'href': u'http://www.freep.com'}, {'seo': u'facebook', 'text': u'facebook', u'href': u'http://www.freep.com/facebook'}, {'seo': u'twitter', 'text': u'twitter', u'href': u'http://www.freep.com/twitter'}],
    "limit_reverse_find": [{'href': 'http://www.freep.com/section/privacy#adchoices', 'seo': 'privacy#adchoices', 'text': 'Ad Choices', 'style': 'color:#4795CC;', 'target': '_blank'}, {'href': 'http://www.freep.com/section/privacy', 'seo': 'privacy', 'text': 'Privacy Notice/Your California Privacy Rights', 'style': 'color:#4795CC;', 'target': '_blank'}, {'href': 'http://www.freep.com/terms', 'seo': 'terms', 'text': 'Terms of Service', 'style': 'color:#4795CC;', 'target': '_blank'}, {'href': 'http://www.freep.com/section/subscription-options&gps-source=CPNLETSUB3', 'seo': 'subscription options&gps source=CPNLETSUB3', 'text': 'Subscribe to the Paper', 'style': 'color:#4795CC;', 'target': '_blank'}, {'href': '%%=RedirectTo(@fullurl)=%%', 'seo': '%%=RedirectTo(@fullurl)=%%', 'text': 'Unsubscribe to electronic mail', 'style': 'color:#4795CC;', 'target': '_blank'}],
    "limit_sort_text": [{'seo': u'm.freep.com', 'text': u'click here', u'style': u'color:#D19D4E;', u'href': u'http://m.freep.com'}, {'seo': u'%%=RedirectTo(@fullurl1)=%%', 'text': u'click here', u'style': u'color:#D19D4E;', u'href': u'%%=RedirectTo(@fullurl1)=%%', u'target': u'_blank'}, {'seo': u'facebook', 'text': u'facebook', u'href': u'http://www.freep.com/facebook'}, {'seo': u'twitter', 'text': u'twitter', u'href': u'http://www.freep.com/twitter'}, {'seo': u'www.freep.com', 'text': u'www.freep.com', u'href': u'http://www.freep.com'}],
    "limit_sort_href": [{'seo': u'%%=RedirectTo(@fullurl1)=%%', 'text': u'click here', u'style': u'color:#D19D4E;', u'href': u'%%=RedirectTo(@fullurl1)=%%', u'target': u'_blank'}, {'seo': u'm.freep.com', 'text': u'click here', u'style': u'color:#D19D4E;', u'href': u'http://m.freep.com'}, {'seo': u'www.freep.com', 'text': u'www.freep.com', u'href': u'http://www.freep.com'}, {'seo': u'facebook', 'text': u'facebook', u'href': u'http://www.freep.com/facebook'}, {'seo': u'twitter', 'text': u'twitter', u'href': u'http://www.freep.com/twitter'}],
    "exclude_links": 144,
    "duplicate_links": 93
}]



""" Unit test Links functionality"""

import re
import os.path
import unittest

import six
import vcr
import bs4
import requests

from linkGrabber import Links
from linkGrabber.tests import test_data as td

BASE_DIR = os.path.dirname(os.path.realpath(__file__))


class TestScrape(unittest.TestCase):
    """ A set of unit tests for Links """
    def setUp(self):
        """ Activated on start up of class """
        self.url = "http://www.google.com"
        self.bad_url = "www.google.com"
        # grab some example html pages to test
        with vcr.use_cassette(os.path.join(BASE_DIR, 'fixtures', 'vcr_cassettes', 'first_test.yaml')):
            td.pages[0]['text'] = requests.get(td.pages[0]['url']).text

        with vcr.use_cassette(os.path.join(BASE_DIR, 'fixtures', 'vcr_cassettes', 'second_test.yaml')):
            td.pages[1]['text'] = requests.get(td.pages[1]['url']).text

    def assertDictSame(self, expected, actual):
        self.assertEqual(len(list(six.iterkeys(expected))), len(list(six.iterkeys(actual))))
        for ekey, evalue in six.iteritems(expected):
            self.assertIn(ekey, actual)
            self.assertEqual(evalue, actual[ekey])

    def test_url(self):
        """ Validate URL on instance instantiation """
        self.assertRaises(Exception, Links, self.bad_url)

    def test_soup_property(self):
        """ Getting the web page yields correct response"""
        seek = Links(self.url)
        self.assertIsInstance(seek._soup, bs4.BeautifulSoup)

    def test_find_limit_param(self):
        """ How does find() handle the limit property """
        seek = Links(self.url)
        self.assertEqual(len(seek.find(limit=5)), 5)
        self.assertEqual(len(seek.find(limit=1)), 1)

    def test_find_number_of_links(self):
        """ Ensure expected number of links reflects actual number of links """
        for page in td.pages:
            seek = Links(text=page['text'])
            self.assertEqual(len(seek.find()), page['num_links'])

    def test_find_limit(self):
        """ Check that the actual array with a limit matches the test data """
        for page in td.pages:
            seek = Links(text=page['text'])
            actual_list = seek.find(limit=5)
            self.assertEqual(len(actual_list), len(page['limit_find']))
            for i, link in enumerate(actual_list):
                self.assertDictSame(link, page['limit_find'][i])

    def test_find_reverse_sort(self):
        """ Ensure reverse sort sorts before limiting the # of links """
        for page in td.pages:
            seek = Links(text=page['text'])
            actual_list = seek.find(limit=5, reverse=True)
            self.assertEqual(len(actual_list), len(page['limit_reverse_find']))
            for i, link in enumerate(actual_list):
                self.assertDictSame(link, page['limit_reverse_find'][i])

    def test_find_sort_by_text(self):
        """ Sorting by text name produces proper results """
        for page in td.pages:
            seek = Links(text=page['text'])
            actual_list = seek.find(limit=5, sort=lambda key: key['text'])
            self.assertEqual(len(actual_list), len(page['limit_sort_text']))
            for i, link in enumerate(actual_list):
                self.assertDictSame(link, page['limit_sort_text'][i])

    def test_find_sort_by_href(self):
        """ Sorting by href produces proper results """
        for page in td.pages:
            seek = Links(text=page['text'])
            actual_list = seek.find(limit=5, sort=lambda key: key['href'] or "")
            self.assertEqual(len(actual_list), len(page['limit_sort_href']))
            for i, link in enumerate(actual_list):
                self.assertDictSame(link, page['limit_sort_href'][i])

    def test_find_exclude(self):
        """ Determine if excluding links removes the links """
        for page in td.pages:
            seek = Links(text=page['text'])
            actual_list = seek.find(exclude=[{"class": re.compile("gb1")}])
            self.assertEqual(len(actual_list), page['exclude_links'])
            actual_list = seek.find(exclude=[{"class": "gb1"}])
            self.assertEqual(len(actual_list), page['exclude_links'])

    def test_find_duplicates(self):
        """ Determine if removing duplicates works """
        for page in td.pages:
            seek = Links(text=page['text'])
            actual_list = seek.find(duplicates=False)
            self.assertEqual(len(actual_list), page['duplicate_links'])

if __name__ == '__main__':
    unittest.main(verbosity=2)



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

def read(fname):
    with open(fname) as fp:
        content = fp.read()
    return content

setup(
    name='linkGrabber',
    version='0.3.0',
    author='Eric Bower',
    author_email='neurosnap@gmail.com',
    packages=['linkGrabber', 'linkGrabber.tests'],
    scripts=[],
    url='https://github.com/michigan-com/linkGrabber',
    license=read('LICENSE.txt'),
    description='Scrape links from a single web page',
    long_description=read('README.rst') + '\n\n' + read('CHANGES.rst'),
    install_requires=["requests", "beautifulsoup4", "six", "lxml"],
    tests_require=['nosetest', 'vcrpy'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP',
        'License :: OSI Approved :: MIT License'
    ],
    keywords=['linkgrabber', 'beautifulsoup', 'scraper',
                'html', 'parser', 'hyperlinks']
)
